# Library of file manipulation functions 
### -*- coding: utf8 -*-

# All input and output arguments are character strings

# Path nodes are separated by "/", even on Windows platforms.
# The adaptation to the current platform is performed by function
# os.path.normpath().

# splitPath     (platform independent)
# splitRealPath (platform independent)
# ExistsDir     (platform independent)
# ExistsFile    (platform independent)


import os.path as op
import time

def neutralizePath(path):
    "replace Windows backslash with slash"
    return path.replace('\\','/')


def splitPath(path=None):
    " Decomposes file and directory paths into various components:"
    " - fullName is the absolute version of path"
    " - prefix is the path minus the last node (directory or file name)"
    " - localFile is the local file name, if path denotes a file, is '' otherwise"
    " - localDir is the local directory name, if path denotes a directory, is '' otherwise"
    " - extension is the extension of the local name, if path denotes a file, is '' otherwise"
    " Special case: if path is a device name (ex.: 'E:' or 'E:/'), both localDir and localFile are ''."
    " If path is not absolute, the current directory is assumed."
    " path must be platform independent (not Windows compliant). If it isn't, it is neutralized."
    " Example: splitPath('E:\\a\\b\\f.v') == splitPath('E:/a/b/f.v')"
    " Path need not denote an existing path on the current machine"
    " So, potential problem for file names without extension."
    #
    # Examples: (the current directory is 'E:/A/B/C') "
    #   path           fullName,       prefix,     localDir, localFile, extension
    #   ------------   --------------  ----------  --------  ---------  ---------
    #   E:/a/b/f.v --> 'E:/a/b/f.v',     'E:/a/b',   '',       'f.v',     'v'      
    #   E:/a/b/c   --> 'E:/a/b/c',       'E:/a/b',   'c',      '',        ''       
    #   E:/f.v     --> 'E:/f.v',         'E:',       '',       'f.v',     'v'      
    #   E:/c       --> 'E:/c',           'E:',       'c',      '',        ''       
    #   E:/        --> 'E:/',            '',         'E:/',    '',        ''       
    #   E:         --> 'E:/',            '',         'E:/',    '',        ''       
    #   d/f.v      --> 'E:/A/B/C/d/f.v', 'E:/A/B/d', '',       'f.v',     'v'      
    #   d/e        --> 'E:/A/B/C/d/e',   'E:/A/B/d', 'e',      '',        ''       
    #   f.v        --> 'E:/A/B/C/f.v',   'E:/A/B',   '',       'f.v',     'v'      
    #   c          --> 'E:/A/B/C/c',     'E:/A/B',   'c',      '',        ''       
    #   .          --> 'E:/A/B/C',       'E:/A/B',   'C',      '',        ''       
    #   ..         --> 'E:/A/B',         'E:/A',     'B',      '',        ''       

    if path == None:
        return u'',u'',u'',u'',u''

    if path == u'':
        return u'',u'',u'',u'',u''

    if path[-1] == ':':                     # 'E:' is not allowed. Transformed into 'E:/'
        path = path + u'/'

    try:
        path = op.normpath(path)            # Normalize for the current platform
    except:
        # Not a valid path
        return u'',u'',u'',u'',u''

    try:
        fullName = op.abspath(path)         # Make the path absolute if needed
    except:
        # Not a valid path
        return u'',u'',u'',u'',u''

    fullName = neutralizePath(fullName)     # Now replace Windows backslash if needed

    # "fullName" is a neutral, absolute path. Let's decompose it.

    prefix,lastNode = op.split(fullName)    # split path by removing the last node
                                            # lastNode is either a file name or a directory name

    localFile = u''
    localDir  = u''
    extension = u''
    
    root,dummy = op.splitdrive(fullName)
    if root+u'/' == fullName:
        # "fullName" = drive name
        prefix = ''
        localDir = fullName

    elif op.isdir(fullName):
        # "fullName" denotes an existing directory
        localDir = lastNode
        
    elif op.isfile(fullName):
        # "fullName" denotes an existing file
        localFile = lastNode
        pos = lastNode.rfind(u'.')
        if pos > -1:
            extension = lastNode[pos+1:]

    else:
        # "fullName" denotes no existing object. Try analyzing the full path.
        pos = lastNode.rfind(u'.')
        if pos > -1:
            # An extension has been found --> file name (duck typing!)
            localFile = lastNode[:pos]
            extension = lastNode[pos+1:]
        else:
            # If it's not a file, it's a directory
            localDir = lastNode

        if prefix[-1] == u'/':                 # If head is a device node, remove terminal '/'
            prefix = prefix[:-1]

    return fullName,prefix,localDir,localFile,extension


def infoFile(path=None):
    # If "path" denotes an existing file, returns its size,
    # its creation, last update and last access times.
    # Otherwise returns (u'',u'',u'',u'').
    
    if path is None:
        return u'',u'',u'',u''

    if path == u'':
        return u'',u'',u'',u''

    try:
        path = op.normpath(path)            # Normalize for the current platform
        fullName = op.abspath(path)         # Make the path absolute if needed
    except Exception,e:
        # Not a valid path
        print e
        return u'',u'',u'',u''
    
    size  = u''
    cTime = u''
    mTime = u''
    aTime = u''

    if op.isfile(fullName):               # The last node is a local file name
        try:
            size  = unicode(op.getsize(fullName))
            cTime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(op.getctime(fullName)))
            mTime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(op.getmtime(fullName)))
            aTime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(op.getatime(fullName)))
        except Exception,e:
            print e
            pass
    
    return size,cTime,mTime,aTime

