-- Database schema for transaction time with granularity "date" --
-- Event times are computed as the last date + a random increment
-- between, say, 0 and 10.


createOrReplaceDB TDB.db;

write-b -- ------------------------------------------- --;
write   -- CREATE TRANSACTION-TIME TEMPORAL DATABASE   --;
write   --  Small version: 3 projects, 8 employees     --;
write-a -- ------------------------------------------- --;

-- Parameters --
-- ---------- --
-- time ID of future
   set future = 999999;
-- timestamp of future
   set Tfuture = 9999-12-31;

-- starting timestamp
   set startDate = 2017-07-24;
-- max time increment
   set maxIncr = 5;

ask-u startDate,maxIncr = [/bParameters for transaction time generation]
                           Start date:|Max increment:;

-- random increment to compute next "end" timestamp
   set Dincr = random_i(0,$maxIncr$);

-- CLOCK stores the events on the database with a unique time ID and 
-- the (non unique) timestamp 
create table CLOCK(
   TimeID integer not null primary key,
   TTime  char(10) not null);

-- Initialize the clock with an arbitrary timestamp --
insert into CLOCK(TTime) values ('$startDate$');

create table H_PROJECT(
   TITLE    char(20)   not null,
   THEME    char(22)   not null,
   BUDGET   decimal(8) not null,
   start    integer    not null,
   end      integer    not null default $future$,
   primary key (TITLE,start));

create table H_EMPLOYEE(
   CODE     char(5)  not null,
   NAME     char(10) not null,
   SALARY   integer  not null,
   CITY     char(10) not null,
   PROJECT  char(20) not null,
   start    integer  not null,
   end      integer  not null default $future$,
   primary key (CODE,start));

-- Current projects --
-- ---------------- --

create view PROJECT(TITLE,THEME,BUDGET)
as select TITLE,THEME,BUDGET
   from   H_PROJECT
   where  end = $future$;

-- Current employees --
-- ----------------- --

create view EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT)
as select CODE,NAME,SALARY,CITY,PROJECT
   from   H_EMPLOYEE
   where  end = $future$;

-- Same as H_PROJECT where time ID have been replaced with true timestamp --
-- ---------------------------------------------------------------------- --
create view TH_PROJECT(TITLE,THEME,BUDGET,start,end)
as select TITLE,THEME,BUDGET,CS.TTime,CE.TTime
   from   H_PROJECT P,CLOCK CS,CLOCK CE
   where  CS.TimeID = P.start and CE.TimeID = P.end
      union
   select TITLE,THEME,BUDGET,CS.TTime,'$Tfuture$'
   from   H_PROJECT P,CLOCK CS
   where  CS.TimeID = P.start and P.end = $future$;

-- Same as H_EMPLOYEE where time ID have been replaced with true timestamp --
-- ----------------------------------------------------------------------- --
create view TH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start,end)
as select CODE,NAME,SALARY,CITY,PROJECT,CS.TTime,CE.TTime
   from   H_EMPLOYEE E,CLOCK CS,CLOCK CE
   where  CS.TimeID = E.start and CE.TimeID = E.end
      union
   select CODE,NAME,SALARY,CITY,PROJECT,CS.TTime,'$Tfuture$'
   from   H_EMPLOYEE E,CLOCK CS
   where  CS.TimeID = E.start and E.end = $future$;


-- Allow ';' at the end of a query in the body of a trigger --
set ; = ;;

-- ------------------ --
-- PROJECT MANAGEMENT --
-- ------------------ --

-- Start a project --
-- --------------- --
create trigger TRG_PRO_INSERT instead of insert on PROJECT
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_PROJECT where TITLE = new.TITLE) $;$
-- 2. create a time event in CLOCK --
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 3. insert data in H_PROJECT with the last time ID --
   insert into H_PROJECT(TITLE,THEME,BUDGET,start) 
   values (new.TITLE,new.THEME,new.BUDGET,(select max(TimeID) from CLOCK)) $;$
end;

-- Update a project --
-- ---------------- --
create trigger TRG_PRO_UPDATE instead of update on PROJECT
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.THEME = old.THEME and new.BUDGET = old.BUDGET $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. close the current state
   update H_PROJECT
   set    end = (select max(TimeID) from CLOCK) 
   where  TITLE = old.TITLE and end = $future$ $;$
-- 5. insert new current state in H_PROJECT (TITLE preserved) --
   insert into H_PROJECT(TITLE,THEME,BUDGET,start) 
   values (old.TITLE,new.THEME,new.BUDGET,(select max(TimeID) from CLOCK)) $;$
end;

-- Stop a project --
-- -------------- --
create trigger TRG_PRO_DELETE instead of delete on PROJECT
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. There can be no dependent EMPLOYEE entities anymore --
   select raise(ABORT,'dependent EMPLOYEEs still exist')
   where  exists(select * from H_EMPLOYEE
                 where  PROJECT = old.TITLE
                 and    end = $future$) $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. close the current state
   update H_PROJECT
   set    end = (select max(TimeID) from CLOCK) 
   where  TITLE = old.TITLE and end = $future$ $;$
end;


-- ------------------- --
-- EMPLOYEE MANAGEMENT --
-- ------------------- --

-- Create an employee --
-- ------------------ --
create trigger TRG_EMP_INSERT instead of insert on EMPLOYEE
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_EMPLOYEE where CODE = new.CODE) $;$
-- 2. The target entity of an Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from H_PROJECT
                     where  TITLE = new.PROJECT
                     and    end = $future$) $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. insert data in H_EMPLOYEE with the last time ID
   insert into H_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) 
   values (new.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,(select max(TimeID) from CLOCK)) $;$
end;

-- Update an employee --
-- ------------------ --
create trigger TRG_EMP_UPDATE instead of update on EMPLOYEE
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.NAME = old.NAME and new.SALARY = old.SALARY
   and    new.CITY = old.CITY and new.PROJECT = old.PROJECT $;$
-- 3. The target entity of the new Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  new.PROJECT <> old.PROJECT
   and    not exists(select * from H_PROJECT
                     where  TITLE = new.PROJECT
                     and    end = $future$) $;$
-- 4. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 5. close the current state
   update H_EMPLOYEE
   set    end = (select max(TimeID) from CLOCK) 
   where  CODE = old.CODE and end = $future$ $;$
-- 6. insert new current state in H_EMPLOYEE (CODE preserved) --
   insert into H_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) 
   values (old.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,(select max(TimeID) from CLOCK)) $;$
end;

-- delete an employee --
-- ------------------ --
create trigger TRG_EMP_DELETE instead of delete on EMPLOYEE
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 3. close the current state
   update H_EMPLOYEE
   set    end = (select max(TimeID) from CLOCK) 
   where  CODE = old.CODE and end = $future$ $;$
end;

closeDB;


-- ------------------------------------------------------ --
-- LOADING INITIAL STATE OF THE TRANSACTION-TIME DATABASE --
-- ------------------------------------------------------ --

parameter commitmode = transaction;

openDB TDB.db;

insert into PROJECT(TITLE,THEME,BUDGET) values('BIOTECH','Biotechnology',180000);
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('G96','Godin',3400,'Genve','BIOTECH');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('M158','Mercier',2900,'Paris','BIOTECH');
update EMPLOYEE set SALARY = 3300 where CODE = 'G96';
update EMPLOYEE set SALARY = 3000 where CODE = 'M158';
update PROJECT set THEME = 'Genetic engineering', BUDGET = 160000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('A68','Albert',3200,'Toulouse','BIOTECH');
update PROJECT set BUDGET = 120000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('D122','Declercq',3200,'Paris','BIOTECH');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('C45','Carlier',3100,'Lille','BIOTECH');
insert into PROJECT(TITLE,THEME,BUDGET) values('SURVEYOR','Satellite monitoring',310000);
update EMPLOYEE set SALARY = 3700,PROJECT = 'SURVEYOR' where CODE = 'A68';
update EMPLOYEE set SALARY = 3600,PROJECT = 'SURVEYOR' where CODE = 'M158';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('D107','Delecourt',3800,'Grenoble','SURVEYOR');
update EMPLOYEE set SALARY = 3800,PROJECT = 'SURVEYOR' where CODE = 'D122';
update EMPLOYEE set SALARY = 4100,CITY = 'Genve' where CODE = 'D107';
update PROJECT set BUDGET = 375000 where TITLE = 'SURVEYOR';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('N240','Nguyen',3100,'Toulouse','BIOTECH');
update PROJECT set BUDGET = 140000 where TITLE = 'BIOTECH';
insert into PROJECT(TITLE,THEME,BUDGET) values('AGRO-2000','Crop improvement',65000);
update EMPLOYEE set SALARY = 3400,PROJECT = 'AGRO-2000' where CODE = 'C45';
update EMPLOYEE set SALARY = 3900,PROJECT = 'AGRO-2000' where CODE = 'G96';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('A237','Antoine',4000,'Grenoble','AGRO-2000');
update EMPLOYEE set SALARY = 3700,CITY = 'Grenoble',PROJECT = 'SURVEYOR' where CODE = 'N240';
update EMPLOYEE set SALARY = 3900,CITY = 'Paris' where CODE = 'A68';
update EMPLOYEE set CITY = 'Grenoble' where CODE = 'D107';
update EMPLOYEE set SALARY = 3100,PROJECT = 'BIOTECH' where CODE = 'C45';
update EMPLOYEE set SALARY = 4100 where CODE = 'G96';
update EMPLOYEE set SALARY = 3800,PROJECT = 'BIOTECH' where CODE = 'A237';
update PROJECT set BUDGET = 345000 where TITLE = 'SURVEYOR';
update EMPLOYEE set CITY = 'Genve',PROJECT = 'AGRO-2000' where CODE = 'N240';
update EMPLOYEE set PROJECT = 'AGRO-2000' where CODE = 'A237';
update EMPLOYEE set SALARY = 3000,PROJECT = 'BIOTECH' where CODE = 'M158';
update EMPLOYEE set SALARY = 3300,PROJECT = 'BIOTECH' where CODE = 'D107';
update PROJECT set BUDGET = 75000 where TITLE = 'AGRO-2000';
update EMPLOYEE set SALARY = 4100,CITY = 'Toulouse',PROJECT = 'AGRO-2000' where CODE = 'D122';
update EMPLOYEE set SALARY = 3900 where CODE = 'G96';
delete from EMPLOYEE where CODE = 'A68';
delete from PROJECT where TITLE = 'SURVEYOR';
update PROJECT set BUDGET = 82000 where TITLE = 'AGRO-2000';
delete from EMPLOYEE where CODE = 'D122';
delete from EMPLOYEE where CODE = 'N240';
update PROJECT set THEME = 'Biotechnology' where TITLE = 'BIOTECH';

commitDB;

select (select count(distinct CODE) from H_EMPLOYEE) as "EMPLOYEE",
       (select count(*) from H_EMPLOYEE) as "H_EMPLOYEE",
       (select count(distinct TITLE) from H_PROJECT) as "PROJECT",
       (select count(*) from H_PROJECT) as "H_PROJECT";
       
write-b Current projects;
select * from PROJECT    order by TITLE;
write-b Current employees;
select * from EMPLOYEE   order by CODE;

write-b History of projects (abstract timestamps);
select * from H_PROJECT  order by TITLE,start;
write-b History of projects (true timestamps);
select * from TH_PROJECT  order by TITLE,start;

write-b History of employees (abstract timestamps);
select * from H_EMPLOYEE order by CODE,start;
write-b History of employees (true timestamps);
select * from TH_EMPLOYEE order by CODE,start;

write-b Clock events;
select * from CLOCK;

closeDB;

parameter commitmode = default;

write-b End of script;

return;

