-- Database schema for transaction time with granularity "date" --
-- Event times are computed as the last date + a random increment
-- between, say, 0 and 10.

-- J-L Hainaut - May 5, 2019


createOrReplaceDB TDB.db;

write-b -- ------------------------------------------------------- --;
write   -- CREATE TRANSACTION-TIME EVENT-BASED TEMPORAL DATABASE   --;
write   --  Small version: 3 projects, 8 employees                 --;
write-a -- ------------------------------------------------------- --;


-- Parameters --
-- ---------- --
-- time ID of future
   set future = 999999;
-- timestamp of future
   set Tfuture = 9999-12-31;

-- starting timestamp
   set startDate = 2017-07-24;
-- max time increment
   set maxIncr = 5;

ask-u startDate,maxIncr = [/bParameters for transaction time generation]
                           Start date:|Max increment:;

-- random increment to compute next "end" timestamp
   set Dincr = random_i(0,$maxIncr$);

-- CLOCK stores the events on the database with a unique time ID and 
-- the (non unique) timestamp 
create table CLOCK(
   TimeID integer not null primary key,
   TTime  char(10) not null);

-- Initialize the clock with an arbitrary timestamp --
insert into CLOCK(TTime) values ('$startDate$');

create table EH_PROJECT(
   TITLE    char(20)   not null,
   THEME    char(22)   not null,
   BUDGET   decimal(8) not null,
   Time     integer    not null,
   Event    char(6)    not null,
   primary key (TITLE,time));

create table EH_EMPLOYEE(
   CODE     char(5)  not null,
   NAME     char(10) not null,
   SALARY   integer  not null,
   CITY     char(10) not null,
   PROJECT  char(20) not null,
   Time     integer    not null,
   Event     char(6)    not null,
   primary key (CODE,time));

-- Current projects --
-- ---------------- --

create view PROJECT(TITLE,THEME,BUDGET)
as select TITLE,THEME,BUDGET
   from   EH_PROJECT P
   where  P.Event in ('create','update')
   and    not exists(select * from EH_PROJECT
                     where  TITLE = P.TITLE
                     and    time > P.time);

-- Current employees --
-- ----------------- --

create view EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT)
as select CODE,NAME,SALARY,CITY,PROJECT
   from   EH_EMPLOYEE E
   where  E.Event in ('create','update')
   and    not exists(select * from EH_EMPLOYEE
                     where  CODE = E.CODE
                     and    time > E.time);


-- Same as EH_PROJECT where time ID have been replaced with true timestamp --
-- ---------------------------------------------------------------------- --


-- Same as EH_EMPLOYEE where time ID have been replaced with true timestamp --
-- ----------------------------------------------------------------------- --


-- Allow ';' at the end of a query in the body of a trigger --
set ; = ;;

-- ------------------ --
-- PROJECT MANAGEMENT --
-- ------------------ --

-- Start a project --
-- --------------- --
create trigger TRG_PRO_INSERT instead of insert on PROJECT
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from EH_PROJECT where TITLE = new.TITLE) $;$
-- 2. create a time event in CLOCK --
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 3. insert data in EH_PROJECT with the last time ID --
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event) 
   values (new.TITLE,new.THEME,new.BUDGET,(select max(TimeID) from CLOCK),'create') $;$
end;

-- Update a project --
-- ---------------- --
create trigger TRG_PRO_UPDATE instead of update on PROJECT
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.THEME = old.THEME and new.BUDGET = old.BUDGET $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. insert new current state in EH_PROJECT (TITLE preserved) --
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event) values
      (old.TITLE,new.THEME,new.BUDGET,(select max(TimeID) from CLOCK),'update') $;$
end;

-- Stop a project --
-- -------------- --
create trigger TRG_PRO_DELETE instead of delete on PROJECT
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. There can be no dependent EMPLOYEE entities anymore --
   select raise(ABORT,'dependent EMPLOYEEs still exist')
   where  exists(select * from EMPLOYEE
                 where  PROJECT = old.TITLE) $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. insert a closed state
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event)
      select TITLE,THEME,BUDGET,(select max(TimeID) from CLOCK),'delete'
      from   PROJECT where TITLE = old.TITLE $;$
end;


-- ------------------- --
-- EMPLOYEE MANAGEMENT --
-- ------------------- --

-- Create an employee --
-- ------------------ --
create trigger TRG_EMP_INSERT instead of insert on EMPLOYEE
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from EH_EMPLOYEE where CODE = new.CODE) $;$
-- 2. The target entity of an Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from PROJECT
                     where  TITLE = new.PROJECT) $;$
-- 3. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 4. insert data in EH_EMPLOYEE with the last time ID
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event) 
   values (new.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,(select max(TimeID) from CLOCK),'create') $;$
end;

-- Update an employee --
-- ------------------ --
create trigger TRG_EMP_UPDATE instead of update on EMPLOYEE
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.NAME = old.NAME and new.SALARY = old.SALARY
   and    new.CITY = old.CITY and new.PROJECT = old.PROJECT $;$
-- 3. The target entity of the new Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from PROJECT
                     where  TITLE = new.PROJECT) $;$
-- 4. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 6. insert new current state in EH_EMPLOYEE (CODE preserved) --
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event) values
      (old.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,(select max(TimeID) from CLOCK),'update') $;$
end;

-- delete an employee --
-- ------------------ --
create trigger TRG_EMP_DELETE instead of delete on EMPLOYEE
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. create a time event in CLOCK
   insert into CLOCK(TTime) select addToDate(max(TTime),$Dincr$) from CLOCK $;$
-- 3. close the current state
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event)
      select CODE,NAME,SALARY,CITY,PROJECT,(select max(TimeID) from CLOCK),'delete'
      from   EMPLOYEE where CODE = old.CODE $;$
end;

closeDB;


-- ------------------------------------------------------ --
-- LOADING INITIAL STATE OF THE TRANSACTION-TIME DATABASE --
-- ------------------------------------------------------ --

parameter commitmode = transaction;

openDB TDB.db;

insert into PROJECT(TITLE,THEME,BUDGET) values('BIOTECH','Biotechnology',180000);
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('G96','Godin',3400,'Genve','BIOTECH');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('M158','Mercier',2900,'Paris','BIOTECH');
update EMPLOYEE set SALARY = 3300 where CODE = 'G96';
update EMPLOYEE set SALARY = 3000 where CODE = 'M158';
update PROJECT set THEME = 'Genetic engineering', BUDGET = 160000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('A68','Albert',3200,'Toulouse','BIOTECH');
update PROJECT set BUDGET = 120000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('D122','Declercq',3200,'Paris','BIOTECH');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('C45','Carlier',3100,'Lille','BIOTECH');
insert into PROJECT(TITLE,THEME,BUDGET) values('SURVEYOR','Satellite monitoring',310000);
update EMPLOYEE set SALARY = 3700,PROJECT = 'SURVEYOR' where CODE = 'A68';
update EMPLOYEE set SALARY = 3600,PROJECT = 'SURVEYOR' where CODE = 'M158';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('D107','Delecourt',3800,'Grenoble','SURVEYOR');
update EMPLOYEE set SALARY = 3800,PROJECT = 'SURVEYOR' where CODE = 'D122';
update EMPLOYEE set SALARY = 4100,CITY = 'Genve' where CODE = 'D107';
update PROJECT set BUDGET = 375000 where TITLE = 'SURVEYOR';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('N240','Nguyen',3100,'Toulouse','BIOTECH');
update PROJECT set BUDGET = 140000 where TITLE = 'BIOTECH';
insert into PROJECT(TITLE,THEME,BUDGET) values('AGRO-2000','Crop improvement',65000);
update EMPLOYEE set SALARY = 3400,PROJECT = 'AGRO-2000' where CODE = 'C45';
update EMPLOYEE set SALARY = 3900,PROJECT = 'AGRO-2000' where CODE = 'G96';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT) values ('A237','Antoine',4000,'Grenoble','AGRO-2000');
update EMPLOYEE set SALARY = 3700,CITY = 'Grenoble',PROJECT = 'SURVEYOR' where CODE = 'N240';
update EMPLOYEE set SALARY = 3900,CITY = 'Paris' where CODE = 'A68';
update EMPLOYEE set CITY = 'Grenoble' where CODE = 'D107';
update EMPLOYEE set SALARY = 3100,PROJECT = 'BIOTECH' where CODE = 'C45';
update EMPLOYEE set SALARY = 4100 where CODE = 'G96';
update EMPLOYEE set SALARY = 3800,PROJECT = 'BIOTECH' where CODE = 'A237';
update PROJECT set BUDGET = 345000 where TITLE = 'SURVEYOR';
update EMPLOYEE set CITY = 'Genve',PROJECT = 'AGRO-2000' where CODE = 'N240';
update EMPLOYEE set PROJECT = 'AGRO-2000' where CODE = 'A237';
update EMPLOYEE set SALARY = 3000,PROJECT = 'BIOTECH' where CODE = 'M158';
update EMPLOYEE set SALARY = 3300,PROJECT = 'BIOTECH' where CODE = 'D107';
update PROJECT set BUDGET = 75000 where TITLE = 'AGRO-2000';
update EMPLOYEE set SALARY = 4100,CITY = 'Toulouse',PROJECT = 'AGRO-2000' where CODE = 'D122';
update EMPLOYEE set SALARY = 3900 where CODE = 'G96';
delete from EMPLOYEE where CODE = 'A68';
delete from PROJECT where TITLE = 'SURVEYOR';
update PROJECT set BUDGET = 82000 where TITLE = 'AGRO-2000';
delete from EMPLOYEE where CODE = 'D122';
delete from EMPLOYEE where CODE = 'N240';
update PROJECT set THEME = 'Biotechnology' where TITLE = 'BIOTECH';

commitDB;


-- Rebuild H_PROJECT from EH_PROJECT  --
-- ---------------------------------- --

create view VEH_PROJECT(TITLE,THEME,BUDGET,start,end) as
  -- Generate all past states --
  select P1.TITLE,P1.THEME,P1.BUDGET,P1.Time,P2.Time
  from   EH_PROJECT P1,EH_PROJECT P2
  where  P1.TITLE = P2.TITLE and P1.Time < P2.time
  and    P1.Event in ('create','update')
  and    not exists(select * from EH_PROJECT
                    where  TITLE = P1.TITLE
                    and    Time > P1.Time and Time < P2.Time)
     union
  -- Generate current states --
  select P.TITLE,P.THEME,P.BUDGET,P.Time,999999
  from   EH_PROJECT P
  where  P.Event in ('create','update')
  and    not exists(select * from EH_PROJECT
                    where  TITLE = P.TITLE
                    and    Time > P.Time)
;

-- Rebuild H_EMPLOYEE from EH_EMPLOYEE --
-- ----------------------------------- --

create view VEH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start,end) as
  -- Generate all past states --
  select P1.CODE,P1.NAME,P1.SALARY,P1.CITY,P1.PROJECT,P1.Time,P2.Time
  from   EH_EMPLOYEE P1,EH_EMPLOYEE P2
  where  P1.CODE = P2.CODE and P1.Time < P2.Time
  and    P1.Event in ('create','update')
  and    not exists(select * from EH_EMPLOYEE
                    where  CODE = P1.CODE
                    and    Time > P1.Time and Time < P2.Time)
     union
  -- Generate current states --
  select P.CODE,P.NAME,P.SALARY,P.CITY,P.PROJECT,P.Time,999999
  from   EH_EMPLOYEE P
  where  P.Event in ('create','update')
  and    not exists(select * from EH_EMPLOYEE
                    where  CODE = P.CODE
                    and    Time > P.Time)
;


write-b -- -------------------------------------------------------- --;
write   -- Event-based history                                      --;
write   -- -------------------------------------------------------- --;

write-b --  EH_PROJECT: event history of projects    --;
select * from EH_PROJECT  order by TITLE,Time;

write-b --  EH_EMPLOYEE: event history of employees   --;
select * from EH_EMPLOYEE order by CODE,Time;

write-b --  CLOCK: clock events                 --;
select * from CLOCK;

write-b --  PROJECT: current projects             --;
select * from PROJECT order by TITLE;

write-b --  EMPLOYEE: current employees            --;
select * from EMPLOYEE order by CODE;


write;
write-b -- -------------------------------------------------------- --;
write   -- Rebuilding entity-based history from event-based history --;
write   -- -------------------------------------------------------- --;

write-b --  VEH_PROJECT: entity history of projects   --;
select * from VEH_PROJECT  order by TITLE,start;

write-b --  VEH_EMPLOYEE: entity history of employees  --;
select * from VEH_EMPLOYEE order by CODE,start;


write;
write-b -- -------------------------------------------------------- --;
write   -- Deriving event-based history from entity-based history   --;
write   -- -------------------------------------------------------- --;

write-b --  Event history of projects   --;
   select TITLE,THEME,BUDGET,start as Time,'create' as Event
   from   VEH_PROJECT P
   where not exists(select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   end = P.start)
      union
   select TITLE,THEME,BUDGET,end as Time,'delete' as Event
   from   VEH_PROJECT P
   where not exists(select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   start = P.end)
   and   end <> $future$
      union
   select TITLE,THEME,BUDGET,start as Time,'update' as Event
   from   VEH_PROJECT P
   where  exists(   select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   end = P.start)
  order by TITLE,Time;


write-b --  Event history of employees   --;
   select CODE,NAME,SALARY,CITY,PROJECT,start as Time,'create' as Event
   from   VEH_EMPLOYEE E
   where not exists(select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   end = E.start)
      union
   select CODE,NAME,SALARY,CITY,PROJECT,end as Time,'delete' as Event
   from   VEH_EMPLOYEE E
   where not exists(select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   start = E.end)
   and   end <> $future$
      union
   select CODE,NAME,SALARY,CITY,PROJECT,start as Time,'update' as Event
   from   VEH_EMPLOYEE E
   where  exists(   select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   end = E.start)
   order by CODE,Time;

closeDB;

parameter commitmode = default;

write-b End of script;

return;

