-- Database schema for valid time with granularity "date" --
-- Event times (either start or end) are provided by the user.

createOrReplaceDB TDB.db;

write-b -- ------------------------------------------- --;
write   -- CREATE VALID-TIME TEMPORAL DATABASE         --;
write   --  Small version: 3 projects, 8 employees     --;
write-a -- ------------------------------------------- --;

-- Parameters --
-- ---------- --
-- timestamp of future
   set Tfuture = 9999-12-31;

create table H_PROJECT(
   TITLE    char(20)   not null,
   THEME    char(22)   not null,
   BUDGET   decimal(8) not null,
   start    date       not null,
   end      date       not null default '$Tfuture$',
   primary key (TITLE,start));

create table H_EMPLOYEE(
   CODE     char(5)  not null,
   NAME     char(10) not null,
   SALARY   integer  not null,
   CITY     char(10) not null,
   PROJECT  char(20) not null,
   start    date     not null,
   end      date     not null default '$Tfuture$',
   primary key (CODE,start));


-- Current projects --
-- --------------- --
create view PROJECT(TITLE,THEME,BUDGET,start,end)
as select TITLE,THEME,BUDGET,start,end
   from   H_PROJECT
   where  end = '$Tfuture$';

-- Current employees --
-- ---------------- --
create view EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start,end)
as select CODE,NAME,SALARY,CITY,PROJECT,start,end
   from   H_EMPLOYEE
   where  end = '$Tfuture$';


-- Allow ';' at the end of a query in the body of a trigger --
set ; = ;;

-- ------------------ --
-- PROJECT MANAGEMENT --
-- ------------------ --

-- Start a project --
-- --------------- --
create trigger TRG_PRO_INSERT instead of insert on PROJECT
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_PROJECT where TITLE = new.TITLE) $;$
-- 2. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start) $;$
-- 3. insert data in H_PROJECT --
   insert into H_PROJECT(TITLE,start,THEME,BUDGET) 
   values (new.TITLE,new.start,new.THEME,new.BUDGET) $;$
end;

-- Update a project --
-- ---------------- --
create trigger TRG_PRO_UPDATE instead of update on PROJECT
when new.end = '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.THEME = old.THEME and new.BUDGET = old.BUDGET $;$
-- 3. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start)
   -- also, must be "monotonic"
   or     new.start <= (select start from H_PROJECT
                        where  TITLE = old.TITLE
                        and    end = '$Tfuture$') $;$
-- 4. close the current state --
   update H_PROJECT
   set    end = new.start 
   where  TITLE = old.TITLE and end = '$Tfuture$' $;$
-- 5. insert new current state in H_PROJECT (TITLE preserved) --
   insert into H_PROJECT(TITLE,start,THEME,BUDGET) 
   values (old.TITLE,new.start,new.THEME,new.BUDGET) $;$
end;

-- Stop a project --
-- -------------- --
create trigger TRG_PRO_DELETE instead of update on PROJECT
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from H_PROJECT
                      where  TITLE = old.TITLE
                      and    end = '$Tfuture$') $;$
-- 3. No more employees on this project --
   select raise(ABORT,'dependent EMPLOYEEs still exist')
   where  exists(select * from H_EMPLOYEE
                 where  PROJECT = old.TITLE and end = '$Tfuture$') $;$
-- 4. close the current state --
   update H_PROJECT
   set    end = new.end 
   where  TITLE = old.TITLE and end = '$Tfuture$' $;$
end;


-- ------------------- --
-- EMPLOYEE MANAGEMENT --
-- ------------------- --

-- Create an employee --
-- ------------------ --
create trigger TRG_EMP_INSERT instead of insert on EMPLOYEE
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_EMPLOYEE where CODE = new.CODE) $;$
-- 2. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start) $;$
-- 3. The target PROJECT must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from H_PROJECT
                     where  TITLE = new.PROJECT
                     and    end = '$Tfuture$') $;$
-- 4. insert data in H_EMPLOYEE with the last time ID --
   insert into H_EMPLOYEE(CODE,start,NAME,SALARY,CITY,PROJECT) 
   values (new.CODE,new.start,new.NAME,new.SALARY,new.CITY,new.PROJECT) $;$
end;

-- Update an employee --
-- ------------------ --
create trigger TRG_EMP_UPDATE instead of update on EMPLOYEE
when new.end = '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.NAME = old.NAME and new.SALARY = old.SALARY
   and    new.CITY = old.CITY and new.PROJECT = old.PROJECT $;$
-- 3. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start)
   -- also, must be "monotonic"
   or     new.start <= (select start from H_EMPLOYEE
                        where  CODE = old.CODE
                        and    end = '$Tfuture$') $;$
-- 4. The new target PROJECT must exist since "new.start" --
   select raise(ABORT,'target PROJECT does not exist')
   where  new.PROJECT <> old.PROJECT
   and    not (    exists(select * from H_PROJECT
                          where  TITLE = new.PROJECT
                          and    end = '$Tfuture$')
               and new.start >= (select min(start) from H_PROJECT
                                 where  TITLE = new.PROJECT)
         ) $;$
-- 5. close the current state --
   update H_EMPLOYEE
   set    end = new.start 
   where  CODE = old.CODE and end = '$Tfuture$' $;$
-- 6. insert new current state in H_EMPLOYEE (CODE preserved) --
   insert into H_EMPLOYEE(CODE,start,NAME,SALARY,CITY,PROJECT) 
   values (old.CODE,new.start,new.NAME,new.SALARY,new.CITY,new.PROJECT) $;$
end;

-- delete an employee --
-- ------------------ --
create trigger TRG_EMP_DELETE instead of update on EMPLOYEE
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from H_EMPLOYEE
                      where  CODE = old.CODE
                      and    end = '$Tfuture$') $;$
-- 3. close the current state --
   update H_EMPLOYEE
   set    end = new.end 
   where  CODE = old.CODE and end = '$Tfuture$' $;$
end;

closeDB;


-- ------------------------------------------------------ --
-- LOADING INITIAL STATE OF THE VALID-TIME DATABASE       --
-- ------------------------------------------------------ --

onError stop;
parameter commitmode = autocommit;

openDB TDB.db;

insert into PROJECT(TITLE,THEME,BUDGET,start) values('BIOTECH','Biotechnology',180000,'2014-11-18');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('G96','Godin',3400,'Genve','BIOTECH','2014-11-23');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('M158','Mercier',2900,'Paris','BIOTECH','2014-12-04');
update EMPLOYEE set start = '2015-01-17',SALARY = 3300 where CODE = 'G96';
update EMPLOYEE set start = '2015-02-21',SALARY = 3000 where CODE = 'M158';
update PROJECT set start = '2015-02-27',THEME = 'Genetic engineering', BUDGET = 160000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('A68','Albert',3200,'Toulouse','BIOTECH','2015-06-09');
update PROJECT set start = '2015-06-15',BUDGET = 120000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('D122','Declercq',3200,'Paris','BIOTECH','2015-06-15');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('C45','Carlier',3100,'Lille','BIOTECH','2015-08-22');
insert into PROJECT(TITLE,THEME,BUDGET,start) values('SURVEYOR','Satellite monitoring',310000,'2015-10-30');
update EMPLOYEE set start = '2015-10-30',SALARY = 3700,PROJECT = 'SURVEYOR' where CODE = 'A68';
update EMPLOYEE set start = '2015-10-30',SALARY = 3600,PROJECT = 'SURVEYOR' where CODE = 'M158';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('D107','Delecourt',3800,'Grenoble','SURVEYOR','2016-02-16');
update EMPLOYEE set start = '2016-03-21',SALARY = 3800,PROJECT = 'SURVEYOR' where CODE = 'D122';
update EMPLOYEE set start = '2016-03-22',SALARY = 4100,CITY = 'Genve' where CODE = 'D107';
update PROJECT set start = '2016-06-25',BUDGET = 375000 where TITLE = 'SURVEYOR';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('N240','Nguyen',3100,'Toulouse','BIOTECH','2016-08-18');
update PROJECT set start = '2016-08-23',BUDGET = 140000 where TITLE = 'BIOTECH';
insert into PROJECT(TITLE,THEME,BUDGET,start) values('AGRO-2000','Crop improvement',65000,'2016-08-31');
update EMPLOYEE set start = '2016-08-31',SALARY = 3400,PROJECT = 'AGRO-2000' where CODE = 'C45';
update EMPLOYEE set start = '2016-08-31',SALARY = 3900,PROJECT = 'AGRO-2000' where CODE = 'G96';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('A237','Antoine',4000,'Grenoble','AGRO-2000','2016-09-16');
update EMPLOYEE set start = '2016-09-16',SALARY = 3700,CITY = 'Grenoble',PROJECT = 'SURVEYOR' where CODE = 'N240';
update EMPLOYEE set start = '2016-10-22',SALARY = 3900,CITY = 'Paris' where CODE = 'A68';
update EMPLOYEE set start = '2016-12-08',CITY = 'Grenoble' where CODE = 'D107';
update EMPLOYEE set start = '2016-12-11',SALARY = 3100,PROJECT = 'BIOTECH' where CODE = 'C45';
update EMPLOYEE set start = '2016-12-11',SALARY = 4100 where CODE = 'G96';
update EMPLOYEE set start = '2016-12-11',SALARY = 3800,PROJECT = 'BIOTECH' where CODE = 'A237';
update PROJECT set start = '2016-12-20',BUDGET = 345000 where TITLE = 'SURVEYOR';
update EMPLOYEE set start = '2016-12-29',CITY = 'Genve',PROJECT = 'AGRO-2000' where CODE = 'N240';
update EMPLOYEE set start = '2016-12-29',PROJECT = 'AGRO-2000' where CODE = 'A237';
update EMPLOYEE set start = '2017-01-14',SALARY = 3000,PROJECT = 'BIOTECH' where CODE = 'M158';
update EMPLOYEE set start = '2017-04-03',SALARY = 3300,PROJECT = 'BIOTECH' where CODE = 'D107';
update PROJECT set start = '2017-04-14',BUDGET = 75000 where TITLE = 'AGRO-2000';
update EMPLOYEE set start = '2017-05-19',SALARY = 4100,CITY = 'Toulouse',PROJECT = 'AGRO-2000' where CODE = 'D122';
update EMPLOYEE set start = '2017-05-19',SALARY = 3900 where CODE = 'G96';
-- delete EMPLOYEE A68 --
update EMPLOYEE set end = '2017-05-19' where CODE = 'A68';
-- delete PROJECT SURVEYOR --
update PROJECT  set end = '2017-05-19' where TITLE = 'SURVEYOR';
update PROJECT set start = '2017-07-05',BUDGET = 82000 where TITLE = 'AGRO-2000';
-- delete EMPLOYEE D112 --
update EMPLOYEE set end = '2017-07-24' where CODE = 'D122';
-- delete EMPLOYEE N240 --
update EMPLOYEE set end = '2017-08-09' where CODE = 'N240';
update PROJECT set start = '2017-09-17',THEME = 'Biotechnology' where TITLE = 'BIOTECH';

commitDB;

select (select count(distinct CODE) from H_EMPLOYEE) as "EMPLOYEE",
       (select count(*) from H_EMPLOYEE) as "H_EMPLOYEE",
       (select count(distinct TITLE) from H_PROJECT) as "PROJECT",
       (select count(*) from H_PROJECT) as "H_PROJECT";

write-b Current projects;
select * from PROJECT  order by TITLE;
write-b Current employees;
select * from EMPLOYEE order by CODE;

write-b History of projects;
select * from H_PROJECT  order by TITLE,start;
write-b History of employees;
select * from H_EMPLOYEE order by CODE,start;

closeDB;

parameter commitmode = default;

write-b End of script;

return;
