
-- ------------------------------ --
-- Computing transitive reduction --
-- ------------------------------ --

-- In table T, each A value has an associated set of B values  --
-- Retrieve all the couples (a1,a2) of A values such that      --
-- the B set of a2 is included into the B set of a1.           --

-- Reduce dataset to check equality relation 

createDB InMemory;
create table T(A char(10), B char(10), primary key (A,B));
insert into T values
  ('01','b'),('01','c'),('01','a'),('01','d'),
  ('02','c'),('02','a'),('02','b'),('02','d'),
  ('03','a'),('03','b'),('03','d'),('03','c'),
  ('04','d'),('04','e'),('04','a'),('04','b'),('04','c'),
  ('05','a'),('05','b'),('05','c'),('05','d'),('05','e'),
  ('06','a'),('06','b'),('06','c'),
  ('07','b'),('07','a');

-- +----+-----------+
-- | A  | BSet      |
-- +----+-----------+
-- | 01 | a,b,c,d   |
-- | 02 | a,b,c,d   |
-- | 03 | a,b,c,d   |
-- | 04 | a,b,c,d,e |
-- | 05 | a,b,c,d,e |
-- | 06 | a,b,c     |
-- | 07 | a,b       |
-- +----+-----------+

-- [1]<----->[2]<---[4]<--->[5]
--  ^         ^
--  |         |
--  +-->[3]<--+
--       |
--       |
--       V
--      [6]--->[7]

-- ----------------------------------------------------- --

extract TS = select count(*) from T;
extract AS = select count(distinct A) from T;
write-b Size of table T  = $TS$ rows;
write Number of A values = $AS$;

create table ABSET as 
       select A, group_concat(B,',') as BSet from T group by A;

create table INCLUSION as
       select C1.A as A1,C2.A as A2
       from   ABSET C1, ABSET C2
       where  C1.A <> C2.A
       and    itemInclude(C1.BSet,C2.BSet,',');

write-b -- Table T --;
select * from T order by A,B;

write-b -- ABSET (order by A) --;
select * from ABSET order by A;

write-b -- ABSET (order by BSet) --;
select * from ABSET order by BSet;

write-b -- INCLUSION (transitive closure) --;
select A1,'includes',A2 from INCLUSION order by A1,A2;

write-b -- INCLUSION (group by A1) --;
select A1,'includes',group_concat(A2,',') as A2 from INCLUSION
group by A1 order by A1,A2;

write-b -- INCLUSION (group by A2) --;
select group_concat(A1,',') as A1,'include', A2 from INCLUSION
group by A2 order by A2;

-- B sets common to A sets --
create view BSET_ASET(BSet,Size,ASet) as
       select BSet, count(*), group_concat(A,'-')
       from   ABSET group by BSet;
write-b -- BSET_ASET (B sets common to A sets) --;
select BSet,Size,ASet as "BSet of" from BSET_ASET

-- B sets common to (not singleton) A sets --
create table EQUAL as 
       select BSet as BSet,group_concat(A,'-') as ASet
       from   ABSET group by BSet having count(*) > 1;
write-b -- EQUAL (B sets common to (not singleton) A sets) --;
select * from EQUAL;

-- Assigning a "group name" to each A --
create view DICTIONARY(A,Replace) as
select ABSET.A, case when BSET_ASET.Size > 1 then BSET_ASET.ASet else ABSET.A end
from   BSET_ASET, ABSET
where  BSET_ASET.BSet = ABSET.BSet;
write-b -- DICTIONARY (assigning a "group name" to each A) --;
select A,Replace as Substitute from DICTIONARY;

write-b -- GR_INCLUSION (= INCLUSION where A1, A2 replaced by "group names") --;
create table GR_INCLUSION as
   select distinct D1.Replace as A1, D2.Replace as A2
   from   INCLUSION I,DICTIONARY D1, DICTIONARY D2
   where  D1.A = I.A1 and D2.A = I.A2
   and    D1.Replace <> D2.Replace;
select * from GR_INCLUSION order by A1,A2;

write-b -- Transitive reduction of GR_INCLUSION --;
select A1,A2
from   GR_INCLUSION I1
where  not exists (select *
                   from   GR_INCLUSION I2, GR_INCLUSION I3
                   where  I2.A1 = I1.A1
                   and    I2.A2 = I3.A1
                   and    I3.A2 = I1.A2)
order by A1,A2;

write-b -- Discarded transitive inclusions --;
select A1,A2
from   GR_INCLUSION I1
where  exists (select *
               from   GR_INCLUSION I2, GR_INCLUSION I3
               where  I2.A1 = I1.A1
               and    I2.A2 = I3.A1
               and    I3.A2 = I1.A2);

-- ----------------------------------------------------------

-- Elementary test with a circuit --

-- [1]<----->[2]
--  |         |
--  |         |
--  +-->[3]<--+

delete from INCLUSION;
insert into INCLUSION
values ('01','02'),('02','01'),('01','03'),('02','03');

select * from INCLUSION;

write-b -- Transitive reduction of INCLUSION with equality --;
select A1,A2
from   INCLUSION I1
where  not exists (select *
                   from   INCLUSION I2, INCLUSION I3
                   where  I2.A1 = I1.A1
                   and    I2.A2 = I3.A1
                   and    I3.A2 = I1.A2)
order by A1,A2;

closeDB;
