# -*- coding: UTF-8 -*-

# Creating proprietary user defined functions

# When the SQLfast engine opens a database, it executes function
# "createFunctions(connect)" of this module
# to attach the UDF defined here below to this database.
# This way, these functions can be used in SQL queries.
# In addition, whether a database is opened or not, these functions
# also are available in SQLfast expressions, in "compute", "if"
# and "while" statements.

# Raising exceptions must be used with caution.
# Their message, whatever their origin (accidental or programmed),
# is replaced by generic string "user-defined function raised exception"
# which provides no information on the exception an is rather useless!
# So, the exception just stops the evaluation.
# To generate a message that better describes the error, save it
# through function "udf.setUDFdiag("<message>")".
# Your message will appear in the following form:
# "Evaluation error ["<message>"], in statement: ..."

# ---------------------------------------

import sqlite3
import UDFdiag as udf   # Used to store informative exception messages


def testException(V):
    # Demonstrates the use of system or manual exceptions in UDF.
    # Returns 1/(V-2) if V <> 2
    # otherwise fires an exception.

    # 1. System exception
    # -------------------
    # Decomment the following to test a system exception
    # temp = V + X

    # 2. Manual exception
    # -------------------
    try:
        temp = 1/(V-2)
    except:
        udf.setUDFdiag(u'function "testException" failed')
        raise Exception(u'')    # Just stops execution
    return temp


def typeOf(V=None):
    # Returns the type of value V
    # types: u'':        undefined (None or unknown type)
    #        u'int':     integer or long
    #        u'decimal': float
    #        u'char':    str or unicode
    # This function is EXCEPTION-FREE

    if isinstance(V,(int,long)):
        return u'int'
    elif isinstance(V,float):
        return u'decimal'
    elif isinstance(V,(str,unicode)):
        return u'char'
    else:
        return u''

def checkNumeric(V=None):
    # Indicates (True,False = 1,0) whether V is a numeric value.
    # N is numeric if it is an integer, a decimal, a real
    # or a numeric character string.
    # This function is EXCEPTION-FREE
    import numbers

    if V is None:
        return False

    if isinstance(V,numbers.Number):
        return True

    if isinstance(V,(str,unicode)):
        # Numeric string?
        # Remove non-digit characters ' -+.'
        res = V.strip(' -+.').replace('.','')
        if res.isdigit():
            return True

    return False

# Experimental functions: Using SQL in an SQLfast UDF
# ===================================================

# A series of (draft) functions showing how to acces a database
# from a UDF in (non SQL) SQLfast statements.
# See SQLfast tutorial, Chapter 22, "Extending SQLfast".

def SQLcustName(cust):
    # Returns the value of "Name" of customer 'cust'.
    conn = sqlite3.connect(u'D:\\SQLfast\\Databases\\ORDERS.db')
    r = conn.execute(u'select Name from CUSTOMER where CustID = "'+cust+u'"').fetchone()
    res = r[0] if r is not None else u''
    conn.close()
    return res

def SQLcustAddress(cust):
    # Returns the value of "Address" of customer 'cust'.
    conn = sqlite3.connect(u'D:\\SQLfast\\Databases\\ORDERS.db')
    r = conn.execute(u'select Address from CUSTOMER where CustID = "'+cust+u'"').fetchone()
    res = r[0] if r is not None else u''
    conn.close()
    return res

def SQLcustCity(cust):
    # Returns the value of "City" of customer 'cust'.
    conn = sqlite3.connect(u'D:\\SQLfast\\Databases\\ORDERS.db')
    r = conn.execute(u'select City from CUSTOMER where CustID = "'+cust+u'"').fetchone()
    res = r[0] if r is not None else u''
    conn.close()
    return res

def SQLcustCat(cust):
    # Returns the value of "Cat" of customer 'cust'.
    conn = sqlite3.connect(u'D:\\SQLfast\\Databases\\ORDERS.db')
    r = conn.execute(u'select Cat from CUSTOMER where CustID = "'+cust+u'"').fetchone()
    res = r[0] if r is not None else u''
    conn.close()
    return res

# ===================================================

def createFunctions(connect):

    try:
        connect.create_function(u"testException",1,testException)
        
        connect.create_function(u"typeOf",1,typeOf)
        connect.create_function(u"checkNumeric", 1, checkNumeric)
        
        connect.create_function(u"custName",1,SQLcustName)
        connect.create_function(u"custAddress",1,SQLcustAddress)
        connect.create_function(u"custCity",1,SQLcustCity)
        connect.create_function(u"custCat",1,SQLcustCat)
        
        return True
    
    except Exception,e:
        print u'[Error in UDF creation ['+e.args[0]+u']. Execution continues]'
        return False

